@@ -1,7 +1,7 @@
 from abc import ABC, abstractmethod
 from copy import deepcopy
 from pathlib import Path
-from math import sqrt
+from math import sqrt, ceil
 import warnings
 import numpy as np
 from scipy import fft, signal
@@ -16,6 +16,7 @@ from bokeh.models.tools import PanTool, BoxZoomTool, WheelZoomTool, ZoomInTool,
 from bokeh.palettes import Viridis256
 from bokeh.io import output_notebook
 from .util import Audio
+from .tf_rep import mel_filterbank, twelve_tet_filterbank, _fft_size_for_spacing
 
 
 try:
@@ -109,16 +110,18 @@ class TimeSignal(Signal):
 
     def __init__(self, in_data, samplerate=None):
         if isinstance(in_data, Spectrogram):
+            if in_data.spacing != 'linear':
+                raise ValueError('Inversion of non-linear spectrograms is not supported')
             if not signal.check_NOLA(in_data.window, in_data._frame_size, in_data._overlap_size):
                 raise ValueError('A spectrogram created with this combination of parameters cannot be inverted')
             if not signal.check_COLA(in_data.window, in_data._frame_size, in_data._overlap_size):
                 warnings.warn('A spectrogram created with this combination of parameters cannot be perfectly inverted')
-            self.times, self.samples = signal.istft(in_data.spectrogram, in_data._samplerate, in_data.window, in_data._frame_size, in_data._overlap_size, in_data._num_bins)
+            self.times, self.samples = signal.istft(in_data.complex_spectrogram, in_data._samplerate, in_data.window, in_data._frame_size, in_data._overlap_size, in_data._fft_size)
             self.samples = np.clip(self.samples, -1, 1)
             self.samplerate = in_data._samplerate
         else:
             if isinstance(in_data, Spectrum):
-                self.samples = fft.irfft(in_data.spectrum, in_data._num_bins, norm='forward')
+                self.samples = fft.irfft(in_data.complex_spectrum, in_data._fft_size, norm='forward')
                 self.samplerate = in_data._samplerate
             else:
                 if samplerate is None:
@@ -246,25 +249,46 @@ class AudioSignal(TimeSignal):
 
 class Spectrum(Signal):
 
-    def __init__(self, in_data, dB=True, num_bins=None, exponent=1, norm_single_side_band=False, samplerate=None):
+    def __init__(
+        self, in_data, *,
+        dB=True, min_freq=None, max_freq=None, num_bins=None,
+        norm_single_side_band=False, spacing='linear', window=None, exponent=1,
+        samplerate=None,
+    ):
         samples, samplerate = get_samples_and_rate(in_data, samplerate)
 
-        if num_bins is None:
-            num_bins = len(samples)
-
         self._samplerate = samplerate
-        self._num_bins = num_bins
+        self.spacing = spacing
+        self.window = window
         self.exponent = exponent
         self._norm_single_side_band = norm_single_side_band
         self.dB = dB
 
-        spectrum = fft.rfft(samples, num_bins, norm='forward')
-        self.magnitude = np.abs(spectrum)
-        self.phase = np.angle(spectrum)
-        self.frequencies = np.arange(len(spectrum)) * samplerate / num_bins
+        if max_freq is None:
+            max_freq = samplerate / 2
+        if min_freq is not None and min_freq < 1 / len(samples):
+            raise ValueError('The given minimum frequency is too small for a signal of this length')
+        if spacing == 'log':
+            if num_bins is None:
+                raise ValueError(f'The number of bins per semitone needs to be given as "num_bins" when requesting {spacing}-spaced frequencies')
+            if min_freq is None or min_freq <= 0 or min_freq >= max_freq:
+                raise ValueError(f'A minimum frequency above 0 and below the maximum frequency needs to be given when requesting {spacing}-spaced frequencies')
+        elif spacing == 'linear':
+            if num_bins is None:
+                num_bins = len(samples)
+        else:
+            raise ValueError('The frequency spacing needs to be one of "linear", "log" or "mel"')
+
+        self._fft_size = max(len(samples), _fft_size_for_spacing(spacing, num_bins, samplerate, min_freq, max_freq))
+        if window is not None:
+            samples = samples * signal.get_window(window, len(samples), True) # do not multiply in place, otherwise in_data gets overwritten
+        spectrum = fft.rfft(samples, self._fft_size, norm='forward')
+        filtered_spectrum, self.frequencies, self._bin_values = _spacing_filter(spectrum, spacing, samplerate, min_freq, max_freq, self._fft_size, num_bins)
+        self.magnitude = np.abs(filtered_spectrum)
+        self.phase = np.angle(filtered_spectrum)
 
         if norm_single_side_band:
-            if self._num_bins % 2 == 0:
+            if self._fft_size % 2 == 0:
                 self.magnitude[1:-1] *= sqrt(2)
             else:
                 self.magnitude[1:] *= sqrt(2)
@@ -275,9 +299,10 @@ class Spectrum(Signal):
 
 
     def plot(self, **fig_args):
+        freq_label = _freq_label(self.spacing)
         default_args = {
             'width': 900, 'height': 300,
-            'x_axis_label': 'frequency [Hz]', 'y_axis_label': 'magnitude',
+            'x_axis_label': freq_label, 'y_axis_label': 'magnitude',
             'tools': [
                 PanTool(),
                 BoxZoomTool(),
@@ -291,8 +316,10 @@ class Spectrum(Signal):
             'active_inspect': 'auto',
             'active_scroll': 'auto',
             'toolbar_location': 'above',
-            'tooltips': [('frequency [Hz]', '@x{0.0}'), ['magnitude', '@y{0.000}']],
+            'tooltips': [('frequency [Hz]', '@frequency{0.000}'), ['magnitude', '@magnitude{0.000}']],
         }
+        if self.spacing in ('mel', 'log'):
+            default_args['tooltips'].insert(0, (freq_label, '@bins{0.0}'))
         if self.exponent == 2 or self.dB:
             default_args['y_axis_label'] = 'power'
             default_args['tooltips'][-1][0] = 'power'
@@ -307,7 +334,8 @@ class Spectrum(Signal):
             fig.x_range.range_padding = 0
         if isinstance(fig.y_range, DataRange1d):
             fig.y_range.range_padding = 0
-        fig.line(self.frequencies, self.magnitude, line_width=2)
+        data_source = ColumnDataSource({'magnitude': self.magnitude, 'frequency': self.frequencies, 'bins': self._bin_values})
+        fig.line(x='bins', y='magnitude', line_width=2, source=data_source)
         return fig
 
 
@@ -326,7 +354,7 @@ class Spectrum(Signal):
             power_per_freq = self.magnitude ** (2/self.exponent)
         if self._norm_single_side_band:
             power = np.sum(power_per_freq)
-        elif self._num_bins % 2 == 0:
+        elif self._fft_size % 2 == 0:
             power = power_per_freq[0] + 2*np.sum(power_per_freq[1:-1]) + power_per_freq[-1]
         else:
             power = power_per_freq[0] + 2*np.sum(power_per_freq[1:])
@@ -354,13 +382,13 @@ class Spectrum(Signal):
 
 
     @property
-    def spectrum(self):
+    def complex_spectrum(self):
         if self.dB:
             magnitude = 10 ** (self.magnitude / 20)
         else:
             magnitude = self.magnitude ** (1/self.exponent)
         if self._norm_single_side_band:
-            if self._num_bins % 2 == 0:
+            if self._fft_size % 2 == 0:
                 self.magnitude[1:-1] /= sqrt(2)
             else:
                 self.magnitude[1:] /= sqrt(2)
@@ -377,21 +405,64 @@ class PowerSpectrum(Spectrum):
 
 class Spectrogram(Signal):
 
-    def __init__(self, input_signal, frame_duration, step_duration, dB=True, num_bins=None, window='hann', exponent=1, samplerate=None):
+    def __init__(
+        self, input_signal, *,
+        frame_duration=None, frame_size=None, step_duration=None, step_size=None, step_ratio=None,
+        dB=True, min_freq=None, max_freq=None, num_bins=None,
+        spacing='linear', window='hann', exponent=1, samplerate=None,
+    ):
         samples, samplerate = get_samples_and_rate(input_signal, samplerate)
 
+        self._num_bins = num_bins
+        self.spacing = spacing
         self.window = window
         self.exponent = exponent
         self.dB = dB
 
         self._samplerate = samplerate
-        self._frame_size = round(frame_duration * samplerate)
-        self._overlap_size = round((frame_duration-step_duration) * samplerate)
-        self._num_bins = num_bins if num_bins is not None else self._frame_size
+        frame_specs = sum([frame_duration is not None, frame_size is not None])
+        if frame_specs != 1:
+            if spacing == 'linear':
+                raise ValueError('Specify a frame width either as duration in seconds or size in samples')
+            elif min_freq is None:
+                raise ValueError('Specify either a minimum frequency or a frame width in seconds or in samples')
+        if frame_duration is not None:
+            self._frame_size = round(frame_duration * samplerate)
+        elif frame_size is not None:
+            self._frame_size = frame_size
+        elif min_freq is not None and spacing != 'linear':
+            self._frame_size = ceil(samplerate / min_freq)
+        if sum([step_duration is not None, step_size is not None, step_ratio is not None]) > 1:
+            raise ValueError('Specify a frame step either as duration in seconds, size in samples or ratio')
+        if step_duration is not None:
+            self._overlap_size = round((frame_duration-step_duration) * samplerate)
+        elif step_size is not None:
+            self._overlap_size = self._frame_size - step_size
+        elif step_ratio is not None:
+            self._overlap_size = round((1-step_ratio) * self._frame_size)
+        else:
+            self._overlap_size = self._frame_size // 2
 
-        self.frequencies, self.times, spectrogram = signal.stft(samples, fs=samplerate, window=window, nperseg=self._frame_size, noverlap=self._overlap_size, nfft=num_bins)
-        self.magnitude = np.abs(spectrogram)
-        self.phase = np.angle(spectrogram)
+        if max_freq is None:
+            max_freq = samplerate / 2
+        if min_freq is not None and min_freq < 1 / self._frame_size:
+            raise ValueError('The given minimum frequency is too small for the given frame width')
+        if spacing == 'log':
+            if num_bins is None:
+                raise ValueError(f'The number of bins per semitone needs to be given as "num_bins" when requesting {spacing}-spaced frequencies')
+            if min_freq is None or min_freq <= 0 or min_freq >= max_freq:
+                raise ValueError(f'A minimum frequency above 0 and below the maximum frequency needs to be given when requesting {spacing}-spaced frequencies')
+        elif spacing == 'linear':
+            if num_bins is None:
+                num_bins = self._frame_size
+        else:
+            raise ValueError('The frequency spacing needs to be one of "linear", "log" or "mel"')
+
+        self._fft_size = max(self._frame_size, _fft_size_for_spacing(spacing, num_bins, samplerate, min_freq, max_freq))
+        _, self.times, spectrogram = signal.stft(samples, fs=samplerate, window=window, nperseg=self._frame_size, noverlap=self._overlap_size, nfft=self._fft_size)
+        filtered_spectrogram, self.frequencies, self._bin_values = _spacing_filter(spectrogram, spacing, samplerate, min_freq, max_freq, self._fft_size, num_bins)
+        self.magnitude = np.abs(filtered_spectrogram)
+        self.phase = np.angle(filtered_spectrogram)
 
         if dB:
             self.magnitude = 20 * np.log10(np.clip(self.magnitude, 1e-6, None))
@@ -407,16 +478,19 @@ class Spectrogram(Signal):
         if not highest_value:
             highest_value = np.max(self.magnitude)
 
+        freq_label = _freq_label(self.spacing)
         default_args = {
             'width': 900, 'height': 400,
-            'x_axis_label': 'time [s]', 'y_axis_label': 'frequency [Hz]',
+            'x_axis_label': 'time [s]', 'y_axis_label': freq_label,
             'tools': 'pan,wheel_zoom,box_zoom,zoom_in,zoom_out,save,reset',
             'active_drag': 'pan',
             'active_inspect': 'auto',
             'active_scroll': 'auto',
             'toolbar_location': 'above',
-            'tooltips': [('time [s]', '@time{0.000}'), ('frequency [Hz]', '@frequency{0.0}'), ['magnitude', '@magnitude{0.000}']],
+            'tooltips': [('time [s]', '@time{0.000}'), ('frequency [Hz]', '@frequency{0.000}'), ['magnitude', '@magnitude{0.000}']],
         }
+        if self.spacing in ('mel', 'log'):
+            default_args['tooltips'].insert(1, (freq_label, '@bins{0.0}'))
 
         if self.exponent == 2 or self.dB:
             default_args['tooltips'][-1][0] = 'power'
@@ -432,23 +506,24 @@ class Spectrogram(Signal):
             fig.x_range.range_padding = 0
         if isinstance(fig.y_range, DataRange1d):
             fig.y_range.range_padding = 0
+        step_time = (self._frame_size - self._overlap_size) / self._samplerate
         if [t for t in fig.tools if isinstance(t, InspectTool)]:
             all_times = np.broadcast_to(self.times, self.magnitude.shape)
             all_freqs = np.broadcast_to(self.frequencies.reshape(-1, 1), self.magnitude.shape)
-            step_time = (self._frame_size - self._overlap_size) / self._samplerate
-            delta_freq = self._samplerate / self._num_bins
-            img_source = ColumnDataSource({'magnitude': self.magnitude.reshape(-1, 1, 1).tolist(), 'time': all_times.ravel() , 'frequency': all_freqs.ravel()})
+            all_bin_values = np.broadcast_to(self._bin_values.reshape(-1, 1), self.magnitude.shape)
+            delta_y = (self._bin_values[-1] - self._bin_values[0]) / (len(self._bin_values) - 1)
+            data_source = ColumnDataSource({'magnitude': self.magnitude.reshape(-1, 1, 1).tolist(), 'time': all_times.ravel(), 'frequency': all_freqs.ravel(), 'bins': all_bin_values.ravel()})
             color_indices = np.rint(np.interp(self.magnitude, (lowest_value, highest_value), (0, len(palette)-1))).astype(int)
-            img_source.data['color'] = [palette[i] for i in color_indices.ravel()]
-            fig.rect(x='time', y='frequency', width=step_time, height=delta_freq, color='color', source=img_source)
+            data_source.data['color'] = [palette[i] for i in color_indices.ravel()]
+            fig.rect(x='time', y='bins', width=step_time, height=delta_y, color='color', source=data_source)
         else:
             mapper = LinearColorMapper(palette=palette, low=lowest_value, high=highest_value)
-            fig.image([self.magnitude], x=self.times[0], y=self.frequencies[0], dw=self.times[-1], dh=self.frequencies[-1], color_mapper=mapper)
+            fig.image([self.magnitude], x=self.times[0]-step_time/2, y=self._bin_values[0], dw=self.times[-1]+step_time, dh=self._bin_values[-1], color_mapper=mapper)
         return fig
 
 
     @property
-    def spectrogram(self):
+    def complex_spectrogram(self):
         if self.dB:
             magnitude = 10 ** (self.magnitude / 20)
         else:
@@ -461,8 +536,39 @@ class Spectrogram(Signal):
             raise ValueError('Specify either the time or the index of the requested spectrum')
         if time is not None:
             index = np.argmin(np.abs(self.times - time))
-        spectrum = Spectrum([0], samplerate=self._samplerate, num_bins=self._num_bins, exponent=self.exponent, dB=self.dB)
+        if self.spacing == 'linear':
+            min_freq, max_freq = None, None
+        else:
+            min_freq, max_freq = self.frequencies[0], self.frequencies[-1]
+        spectrum = Spectrum(np.zeros(self._frame_size), dB=self.dB, min_freq=min_freq, max_freq=max_freq, 
+            num_bins=self._num_bins, norm_single_side_band=False, spacing=self.spacing,
+            window=self.window, exponent=self.exponent, samplerate=self._samplerate)
+        spectrum._fft_size = self._fft_size
         spectrum.magnitude = self.magnitude[:, index]
         spectrum.phase = self.phase[:, index]
         spectrum.frequencies = self.frequencies
         return spectrum
+
+
+def _spacing_filter(spectral_values, spacing, samplerate, min_freq, max_freq, fft_size, num_bins):
+    if spacing == 'linear':
+        lin_frequencies = np.arange(len(spectral_values)) * samplerate / fft_size
+        if min_freq is not None:
+            min_idx = np.argmin(np.abs(np.where(lin_frequencies < min_freq, np.inf, lin_frequencies) - min_freq))
+        if max_freq is not None:
+            max_idx = np.argmin(np.abs(np.where(lin_frequencies > max_freq, np.inf, lin_frequencies) - max_freq))
+        return spectral_values, lin_frequencies, lin_frequencies
+    if spacing == 'mel':
+        bin_values, spaced_frequencies, filterbank = mel_filterbank(samplerate, fft_size, num_bins, min_freq, max_freq, normalize=False)
+    elif spacing == 'log':
+        bin_values, spaced_frequencies, filterbank = twelve_tet_filterbank(samplerate, fft_size, num_bins, min_freq, max_freq, normalize=False, a4=440)
+    return np.matmul(filterbank, spectral_values), spaced_frequencies, bin_values
+
+
+def _freq_label(spacing):
+    if spacing == 'linear':
+        return 'frequency [Hz]'
+    if spacing == 'mel':
+        return 'mel'
+    if spacing == 'log':
+        return 'MIDI number'
